import asyncio
import logging
from base64 import b64decode
from functools import partial
from json import dumps, loads
from typing import Optional, Tuple, Type

from aiohttp import ClientResponse, ClientSession

from .error import (
    BadRequestException,
    ForbiddenException,
    InternalServerErrorException,
    NotFoundException,
    TooManyRequestsException,
    UnauthorizedException,
)
from .guild import Guild, Leaderboard, Sort
from .json_encoder import Encoder as JSONEncoder
from .number import Number
from .perms import Permissions
from .user import User

log = logging.getLogger(__name__)

__all__ = ("Client",)


def no_none(d: dict) -> dict:
    return {i: j for i, j in d.items() if j is not None}


def serialize_values(
    d: dict, serialize=dumps, classes: Tuple[Type, ...] = (str, int, float)
) -> dict:
    return {
        i: j if isinstance(j, classes) else serialize(j) for i, j in d.items()
    }


class Client:
    """An UnbelievaBoat API client. The heart of the Pybelieva."""

    _json_serializer = partial(dumps, cls=JSONEncoder)
    _url: str
    _session: ClientSession
    id: int
    wait_on_ratelimits: bool

    def __init__(
        self,
        token: str,
        *,
        base_url: str = "https://unbelievable.pizza/api",
        version: int = 1,
        wait_on_ratelimits: bool = True,
    ):
        self._url = f"{base_url}/v{version}"
        self._session = ClientSession(
            headers={"Authorization": token},
            json_serialize=self._json_serializer,
        )
        self.id = int(loads(b64decode(token.split(".")[1]).decode())["app_id"])
        self.wait_on_ratelimits = wait_on_ratelimits

    async def close(self):
        await self._session.close()

    def oauth_url(self, guild_id: Optional[int] = None) -> str:
        return (
            "https://unbelievaboat.com/applications/authorize"
            f"?app_id={self.id}"
            + (f"&guild_id={guild_id}" if guild_id else "")
        )

    async def __request(
        self, method: str, path: str, *, data: dict = None, query: dict = None
    ):
        resp: ClientResponse = await self._session.request(
            method,
            f"{self._url}{path}",
            json=data,
            params=serialize_values(query, serialize=self._json_serializer)
            if query is not None
            else None,
        )
        log.debug(f"{resp.request_info.method} {resp.url} {data}")
        if resp.status == 400:
            raise BadRequestException()
        if resp.status == 401:
            raise UnauthorizedException()
        if resp.status == 403:
            raise ForbiddenException()
        if resp.status == 404:
            raise NotFoundException()
        if resp.status == 500:
            raise InternalServerErrorException()
        json: dict = await resp.json()
        log.debug(f"{json=}")
        if resp.status == 429:
            retry_after = json.get("retry_after", 1000)
            log.info(f"Ratelimited for {retry_after}ms")
            if not self.wait_on_ratelimits:
                raise TooManyRequestsException(**json)
            await asyncio.sleep(retry_after / 1000)
            return await self.__request(method, path, data=data, query=query)
        return json

    async def get_user_balance(self, guild_id: int, user_id: int) -> User:
        return User.parse_obj(
            await self.__request("GET", f"/guilds/{guild_id}/users/{user_id}")
        )

    async def set_user_balance(
        self,
        guild_id: int,
        user_id: int,
        *,
        cash: Optional[Number] = None,
        bank: Optional[Number] = None,
        reason: Optional[str] = None,
    ) -> User:
        return User.parse_obj(
            await self.__request(
                "PUT",
                f"/guilds/{guild_id}/users/{user_id}",
                data=no_none({"cash": cash, "bank": bank, "reason": reason}),
            )
        )

    async def patch_user_balance(
        self,
        guild_id: int,
        user_id: int,
        cash: Optional[Number] = None,
        bank: Optional[Number] = None,
        reason: Optional[str] = None,
    ) -> User:
        return User.parse_obj(
            await self.__request(
                "PATCH",
                f"/guilds/{guild_id}/users/{user_id}",
                data=no_none({"cash": cash, "bank": bank, "reason": reason}),
            )
        )

    async def get_guild_leaderboard(
        self,
        guild_id: int,
        *,
        sort: Optional[Sort] = None,
        offset: Optional[int] = None,
        page: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Leaderboard:
        return Leaderboard.parse_obj(
            await self.__request(
                "GET",
                f"/guilds/{guild_id}/users",
                query=no_none(
                    {
                        "offset": offset,
                        "page": page,
                        "limit": limit,
                        "sort": sort,
                    }
                ),
            )
        )

    async def get_guild(self, guild_id: int) -> Guild:
        return Guild(**await self.__request("GET", f"/guilds/{guild_id}"))

    async def get_permissions(self, guild_id: int) -> Permissions:
        return Permissions(
            (
                await self.__request(
                    "GET", f"/applications/@me/guilds/{guild_id}"
                )
            )["permissions"]
        )



from json import JSONEncoder

from .guild import Sort
from .number import Infinity

__all__ = ("Encoder",)


class Encoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Infinity):
            return str(o)
        if isinstance(o, Sort):
            return o.value
        return super().default(o)



import logging

from .client import Client
from .error import (
    BadRequestException,
    ForbiddenException,
    HTTPException,
    InternalServerErrorException,
    NotFoundException,
    PybelievaException,
    TooManyRequestsException,
    UnauthorizedException,
)
from .guild import Guild, Leaderboard, Sort
from .number import (
    INFINITY,
    NEGATIVE_INFINITY,
    POSITIVE_INFINITY,
    Sign,
    Infinity,
    Number,
)
from .perms import Permissions
from .user import User

__all__ = (
    "BadRequestException",
    "Client",
    "ForbiddenException",
    "Guild",
    "HTTPException",
    "INFINITY",
    "Infinity",
    "InternalServerErrorException",
    "Leaderboard",
    "NEGATIVE_INFINITY",
    "NotFoundException",
    "Number",
    "POSITIVE_INFINITY",
    "Permissions",
    "PybelievaException",
    "Sign",
    "Sort",
    "TooManyRequestsException",
    "UnauthorizedException",
    "User",
)
logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybelieva']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6,<4.0', 'pydantic>=1.7,<2.0']

setup_kwargs = {
    'name': 'pybelieva',
    'version': '0.1.0.dev5',
    'description': 'Python wrapper for UnbelievaBoat API',
    'long_description': '# Pybelieva\n\n[![discord][discord-badge]][discord] [![api][api-badge]][api]\n![logo][logo]\n\nPybelieva is a library providing a Python interface\nto the [UnbelievaBoat API][api]. To\nuse it, you need to get a token from the official documentation page.\n\n## Dependencies\n\n- Python\n- aiohttp\n- discord.py (optional)\n\n## Roadmap\n\n- Documentation\n- Rate limit handling\n\n[logo]: https://i.imgur.com/RLRDeQw.png\n[discord-badge]: https://img.shields.io/discord/592830069806989339?color=4e5d94&label=Discord&logo=discord\n[discord]: https://discord.gg/jK8thg37mC\n[api]: https://unb.pizza/api/docs\n[api-badge]: https://img.shields.io/badge/api-v1-ff266a\n',
    'author': 'John Meow',
    'author_email': 'j0hn.meow@ya.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/dev-cats/Pybelieva',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class Permissions:
    _bits: int

    def __and__(self, rhs: Permissions) -> Permissions:
        return Permissions(self._bits & rhs._bits)

    def __or__(self, rhs: Permissions) -> Permissions:
        return Permissions(self._bits | rhs._bits)

    def __xor__(self, rhs: Permissions) -> Permissions:
        return Permissions(self._bits ^ rhs._bits)

    @property
    def economy(self):
        return bool(self._bits & 0x1)
